/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "aperture.hh"
#include "lcl_array.hh"
#include "DVecType.hh"
#include "framedir.hh"
#include "jsonStack.hh"
#include "xsil/MetaIO.hh"
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iostream>

using namespace std;

EXECDAT(aperture)

//======================================  constructor
aperture::aperture(int argc, const char* argv[]) 
: DatEnv(argc, argv), _CL(0.998), _corr_limit(0.5), _trig_total(0),
   _osc(getDacc())
{
   _config = argv[argc-1];
   configure(_config);
   getDacc().setStride(16.0);
}

//======================================  Destructor
aperture::~aperture(void) {
   cout << "Starting post-processing" << endl;
   gps_t start = _channel[0].ref_values().getStartTime().getS();
   gps_t stop  = _channel[0].ref_values().getEndTime().getS();
   Interval dt = _channel[0].ref_values().getTStep();
   size_t nBin = double(stop - start)/double(dt);
   cout << "reading triggers from: " << _trig_path 
        << " gps range: " << start << "-" << stop << endl;
   unique_ptr<DVectI> dvi(new DVectI(nBin));
   dvi->replace_with_zeros(0, nBin, nBin);
   _trig_series.setData(Time(start), dt, dvi.release());
   cout << "Trigger selection criteria: " << endl;
   for (size_t i=0; i<_trig_select_list.size(); i++) {
      _trig_select_list[i].critter(cout);
   }
   read_triggers(_trig_path, start, stop);
   cout << "Vetoing triggers" << endl;
   veto_triggers();
   cout << "Starting channel pair test" << endl;
   calculate();
}

//======================================  Poisson function
//  this function returns the probability that a randomly (poisson) distributed
//  sample with mean x will have >=i events.
static double
poissonge(double x, int i) {
   double res = 1.0;
   if (i > 0) {
      double val = exp(-x);
      res -= val;
      for (size_t n=1; n<i; n++) {
	 val *= x / double(n);
	 res -= val;
      }
   }
   return res;
}

//======================================  Calculate everything.
void
aperture::calculate(void) {
   //-----------------------------------  Get limits perform 1D tests
   size_t N = _channel.size();
   for (size_t i=0; i<N; i++) {
      if (_channel[i].empty()) {
	 cout << _channel[i].name() << " is empty!" << endl;
         continue;
      }
      _channel[i].set_limits();
      _channel[i].test1D(_trig_series, _CL);
   }

   //-----------------------------------  get the expected number per cell.
   size_t M=16;

   cout << "start 2D tests..." << endl;
   lcl_array<int> kArray(M*M);
   lcl_array<int> tArray(M*M);
   const DVectI& dvt = dynamic_cast<DVectI&>(*_trig_series.refDVect());
   for (size_t i=1; i<N; i++) {
      if (_channel[i].reject()) continue;
      DVectI dvi(*_channel[i].ref_rank().refDVect());
      size_t nEnt = dvi.size();
      double trigAvg = double(_trig_total) / double(nEnt);
      for (size_t j=0; j<i; j++) {
         if (_channel[j].reject()) continue;
	 //cout << "computing {" << _channel[i].name() << ", "
	 //     << _channel[j].name() << "}" << endl;
	 
	 //-----------------------------  Zero statistics arrays
	 memset(kArray.get(), 0, M*M*sizeof(int));
	 memset(tArray.get(), 0, M*M*sizeof(int));
	 
	 //-----------------------------  fill statistics
	 DVectI dvj(*_channel[j].ref_rank().refDVect());
	 for (size_t k=0; k<nEnt; k++) {
	    size_t ivsj = dvi[k] * M + dvj[k];
	    kArray[ivsj]++;
	    tArray[ivsj] += dvt[k];
	 }
	 
	 //-----------------------------  flag stats
	 int savij = -1;
         double wx = 1 - _CL;
	 double ssq=0;
	 for (size_t inxij=0; inxij<M*M; inxij++) {
	    ssq += pow(kArray[inxij], 2);
	    double xpect = kArray[inxij]*trigAvg;
            double pois = poissonge(xpect, tArray[inxij]);
	    if (pois < wx) {
               wx = pois;
               savij = inxij;
	    }
	 }
	 double corr = sqrt(M * ssq / (nEnt *nEnt));
         if (savij >= 0 && corr < _corr_limit) {
            size_t kj = savij % M;
            size_t ki = savij / M;
            cout << _channel[i].name() << "[" << ki << "] " 
                 << _channel[j].name() << "[" << kj << "] N=" << tArray[savij]
	         << " expect=" << kArray[savij]*trigAvg 
                 << " pois=" << wx << endl;
	    cout << "correlation = " << corr << endl;
         }
      }
   }
}

//======================================  Read the configuration file
void
aperture::configure(const std::string& cfile) {
   jsonStack js(cfile);

   //-----------------------------------  Numeric parameters
   js.fetch_data("confidence", _CL);
   js.fetch_data("correlation-limit", _corr_limit);

   //-----------------------------------  OSC info
   js.fetch_data("osc-file", _oscfile);
   js.fetch_data("osc-cond", _osc_cond);

   //-----------------------------------  Trigger selections.
   js.fetch_data("trigger-path", _trig_path);
   js.fetch_data("trigger-table-name", _trig_table_name);
   js.fetch_data("trigger-time-field", _trig_time_field);
   js.fetch_data("trigger-time-ns-field", _trig_time_ns_field);
   if (js.push_element(string("trigger-select"))) {
      while (js.iterate()) {
         trig_select ts;
         js.fetch_data("field", ts.field);
         if (js.fetch_data("string-eq", ts.string)) {
	    ts.type = trig_select::s_string;
         } else if (js.fetch_data("num-ge", ts.number)) {
	    ts.type = trig_select::s_num_ge;
         } else {
	    ts.type = trig_select::s_pass;
         }
         _trig_select_list.push_back(ts);
         js.pop();
      }
      js.pop();
   }

   if (js.push_element(string("vetos"))) {
      while (js.iterate()) {
         trig_veto tv;
	 tv.val_min = -1.0/0.0;
	 tv.val_max =  1.0/0.0;
         js.fetch_data("channel", tv.channel);
         js.fetch_data("val-min", tv.val_min);
         js.fetch_data("val-max", tv.val_max);
	 _trig_veto_list.push_back(tv);
         js.pop();
      }
      js.pop();
   }

   //-----------------------------------  Channel list
   std::vector<std::string> chan_vect;
   size_t nChan = js.fetch_data("channels", chan_vect);
   if (!nChan) throw runtime_error("aperture: Error reading channel list");
   for (size_t i=0; i<nChan; i++) {
      _channel.push_back(channels(chan_vect[i]));
      getDacc().addChannel(chan_vect[i]);
   }

   //------------------------------------  Set stuff up
   if (!_oscfile.empty()) _osc.readConfig(_oscfile);

   //------------------------------------  Print
   cout << "confidence:            "   << _CL << endl;
   cout << "osc-file:              \"" << _oscfile << "\"" << endl;
   cout << "osc-cond:              \"" << _osc_cond << "\"" << endl;
   cout << "trigger-path           \"" << _trig_path << "\"" << endl;
   cout << "trigger-table-name     \"" << _trig_table_name << "\"" << endl;
   cout << "trigger-time-field     \"" << _trig_time_field << "\"" << endl;
   cout << "trigger-time-ns-field  \"" << _trig_time_ns_field << "\"" << endl;
   cout << "trigger selections:    "   << _trig_select_list.size() << endl;
   cout << "trigger vetoes:        "   << _trig_veto_list.size() << endl;
   cout << "number of channels:    "   << _channel.size() << endl;
}

//======================================  Collect data from specified channels
void
aperture::ProcessData(void) {
   cout << "read data data at: " << getDacc().getFillTime() << endl;
   if (!_osc_cond.empty() && !_osc.satisfied(_osc_cond)) return;
   size_t N = _channel.size();
   for (size_t i=0; i<N; i++) {
      const TSeries* ts = getDacc().refData(_channel[i].name());
      if (!ts) {
         cerr << "aperture: no data for channel: " << _channel[i].name() <<endl;
         continue;
      }
      _channel[i].append(*ts);
   }
}

//=======================================  Read trigger data
void
aperture::read_triggers(const std::string& file, gps_t start, gps_t stop) {
   size_t nRead(0), nTime(0), nSelect(0);
   FrameDir fd;
   fd.add(file.c_str(), true);
   FrameDir::file_iterator endfile = fd.getLast(stop);
   DVectI& dvi = dynamic_cast<DVectI&>(*_trig_series.refDVect()); 
   for (FrameDir::file_iterator i=fd.getStart(start); i != endfile; ++i) {
      xsil::MetaIO mio(i->getFile().c_str(), _trig_table_name.c_str());
      if (!mio.is_open()) continue;
      while (mio.getRow()) {
	 nRead++;
	 Time::ulong_t gps_s  = mio.getInt(_trig_time_field.c_str());
	 if (gps_s < start || gps_s >= stop) continue;
         nTime++;
	 Time::ulong_t gps_ns = mio.getInt(_trig_time_ns_field.c_str());
	 if (select_trigger(mio)) {
	    nSelect++;
	    size_t ibin = _trig_series.getBin(Time(gps_s, gps_ns));
	    dvi[ibin]++;
            _trig_total++;
	 }
      }
   }
   cout << "triggers read:       " << nRead << endl;
   cout << "triggers in time:    " << nTime << endl;
   cout << "triggers selected:   " << nSelect << endl;
}

//=======================================  Test trigger selection criteria
bool
aperture::select_trigger(const xsil::MetaIO& mio) const {
   size_t N = _trig_select_list.size();
   for (size_t i=0; i<N; i++) {
      const trig_select& tsi = _trig_select_list[i];
      switch (tsi.type) {
      case trig_select::s_string:
	 if (mio.getString(tsi.field.c_str()) != tsi.string) return false;
	 break;
      case trig_select::s_num_ge:
	 if (mio.getFloat(tsi.field.c_str()) < tsi.number) return false;
	 break;
      default:
	 break;
      }
   }
   return true;
}

//=======================================  Veto triggers
void
aperture::veto_triggers(void) {
   size_t nVetoed = 0;
   size_t Nc = _channel.size();
   size_t Nv = _trig_veto_list.size();
   cout << setprecision(3) << fixed;
   for (size_t iv=0; iv<Nv; iv++) {
      int ic = -1;
      for (size_t i=0; i<Nc; i++) {
	 if (_channel[i].name() == _trig_veto_list[iv].channel) {
	    ic = i;
	    break;
	 }
      }
      if (ic < 0) continue;
      double veto_min = _trig_veto_list[iv].val_min;
      double veto_max = _trig_veto_list[iv].val_max;
      cout << "veto triggers with " << veto_min << " <= "
	   << _trig_veto_list[iv].channel  << " < " << veto_max << endl;
      size_t jCol = 0;
      size_t Ne = _channel[ic].ref_values().getNSample();
      DVectI& trigDv(*dynamic_cast<DVectI*>(_trig_series.refDVect()));
      const TSeries& data = _channel[ic].ref_values();
      for (size_t i=0; i<Ne; i++) {
	 if (!trigDv[i]) continue;
	 double x = data.getDouble(i);
	 if (x >= veto_min && x < veto_max) {
	    if (!jCol) cout << "vetoed:";
	    cout << " " << data.getBinT(i).totalS();
	    if (++jCol == 5) {
	       cout << endl;
	       jCol = 0;
	    }
	    nVetoed += trigDv[i];
	    trigDv[i] = 0;
	 }
      }
      if (jCol) cout << endl;
   }
   cout.unsetf(ios::fixed);
   cout << "triggers vetoed:     " << nVetoed << endl;
   _trig_total -= nVetoed;
   cout << "total triggers:      " << _trig_total << endl;
}

//======================================  Print trigger selection rule
void 
aperture::trig_select::critter(std::ostream& out) const {
    out << "field: " << field;
    switch (type) {
    case s_pass:
       out << " pass";
       break;
    case s_string:
       out << " equal to \"" << string;
       break;
    case s_num_ge:
       out << " >= " << number;
       break;
    default:
       out << " undefined operation.";
    }
    out << endl;
}

//======================================  Channel constructor
channels::channels(const std::string& chan, size_t N)
   : _alloc(N), _name(chan), _reject(false)
{
}

//=======================================  Destructor
channels::~channels(void) {
}

//=======================================  Append data
void 
channels::append(const TSeries& ts) {
   if (ts.empty()) return;
   if (_value.empty()) {
      DVector* dv = ts.refDVect()->clone();
      dv->reserve(_alloc);
      _value.setData(ts.getStartTime(), ts.getTStep(), dv);
   } else {
      _value.Append(ts);
   }
}
   
//=======================================  Set the limits
void 
channels::set_limits(size_t M) {
   DVectF cdata(*_value.refDVect());
   DVectF data = cdata;
   size_t N=data.size();
   std::sort(data.refTData(), data.refTData()+N);
   _limits.resize(M+1, 0);

   //-----------------------------------  Pick up the rank limit values.
   _limits[0] = data[0];
   _limits[M] = data[N-1];
   size_t div = N / M;
   size_t ndup = 0;
   for (size_t i=1; i<M; i++) {
      size_t inx = i * div;
      _limits[i] = data[inx];
      if (_limits[i] == data[inx-1]) ndup++;
   }
   if (ndup) {
      cout << "warning - " << ndup << " slices of channel: " << name()
           << " may share values." << endl;
      if (ndup == 15) set_reject(true);
   }

   //-------------------------------------  Set up the rank Series.
   DVectI rvect(N);
   size_t rinx = 0;
   for (size_t i=0; i<N; i++) {
      float cdi = cdata[i];
      while (cdi < _limits[rinx] || cdi >= _limits[rinx+1]) {
	 if (cdi < _limits[rinx]) {
	    if (!rinx) break;
	    rinx--;
	 } else if (cdi >= _limits[rinx+1]) {
	    if (rinx == M-1) break;
	    rinx++;
	 }
      }
      rvect[i] = rinx;
   }
   _rank.setData(_value.getStartTime(), _value.getTStep(), rvect.clone());
}

//=====================================  Test the trigger rates per rank
void
channels::test1D(const TSeries& ts, double CL) {
   size_t N = _rank.getNSample();
   double nTrig  = ts.refDVect()->VSum(0, N);
   double perbin = nTrig/double(N);
   size_t M = 16;

   //----------------------------------  Set up count arrays
   lcl_array<size_t> kval(M);
   lcl_array<size_t> tval(M);
   memset(kval.get(), 0, M*sizeof(size_t));
   memset(tval.get(), 0, M*sizeof(size_t));

   //----------------------------------  Count entries and triggers
   const DVectI& dvr(dynamic_cast<const DVectI&>(*_rank.refDVect()));
   const DVectI& dvt(dynamic_cast<const DVectI&>(*ts.refDVect()));
   for (size_t i=0; i<N; i++) {
      int k = dvr[i];
      kval[k]++;
      tval[k] += dvt[i];
   }

   //----------------------------------  Check the stats
   int wi=-1;
   double wx = 1 - CL;
   double wav; 
   for (size_t i=0; i<M; i++) {
      double average = perbin * kval[i];
      double pois    = poissonge(average, tval[i]);
      if (pois < wx) {
         wi = i;
         wx = pois;
         wav = average;
      }
   }
   if (wi >= 0) {
      cout << name() << "[" << wi << "] N=" << tval[wi] 
           << " expect=" << wav << " pois=" << wx << endl;
      cout << "            range " << wi << ": " << _limits[wi] << " - "
	   << _limits[wi+1] << endl;
      //set_reject(true);
   }
}

//===================================== set the reject bit
void
channels::set_reject(bool yorn) {
   _reject = yorn;
}
